from __future__ import annotations

from typing import ClassVar

import diffrax
from jax import numpy as jnp
from jax import random as jr
from jaxtyping import Array, ArrayLike, Bool, Float, Int, Key

from lerax.render import AbstractRenderer, Color, PygameRenderer, Transform
from lerax.space import Box, Discrete

from .base_classic_control import (
    AbstractClassicControlEnv,
    AbstractClassicControlEnvState,
)


class AcrobotState(AbstractClassicControlEnvState):
    y: Float[Array, "4"]
    t: Float[Array, ""]


class Acrobot(
    AbstractClassicControlEnv[AcrobotState, Int[Array, ""], Float[Array, "4"]]
):
    name: ClassVar[str] = "Acrobot"

    action_space: Discrete
    observation_space: Box

    gravity: Float[Array, ""]
    link_length_1: Float[Array, ""]
    link_length_2: Float[Array, ""]
    link_mass_1: Float[Array, ""]
    link_mass_2: Float[Array, ""]
    link_com_pos_1: Float[Array, ""]
    link_com_pos_2: Float[Array, ""]
    link_moi: Float[Array, ""]
    max_vel_1: Float[Array, ""]
    max_vel_2: Float[Array, ""]
    torque_max_noise: Float[Array, ""]
    torques: Float[Array, "3"]

    dt: Float[Array, ""]
    solver: diffrax.AbstractSolver
    dt0: Float[Array, ""] | None
    stepsize_controller: diffrax.AbstractStepSizeController

    def __init__(
        self,
        *,
        gravity: Float[ArrayLike, ""] = 9.8,
        link_length_1: Float[ArrayLike, ""] = 1.0,
        link_length_2: Float[ArrayLike, ""] = 1.0,
        link_mass_1: Float[ArrayLike, ""] = 1.0,
        link_mass_2: Float[ArrayLike, ""] = 1.0,
        link_com_pos_1: Float[ArrayLike, ""] = 0.5,
        link_com_pos_2: Float[ArrayLike, ""] = 0.5,
        link_moi: Float[ArrayLike, ""] = 1.0,
        max_vel_1: Float[ArrayLike, ""] = 4 * jnp.pi,
        max_vel_2: Float[ArrayLike, ""] = 9 * jnp.pi,
        torque_max_noise: Float[ArrayLike, ""] = 0.0,
        torques: Float[ArrayLike, "3"] = jnp.array([-1.0, 0.0, 1.0]),
        dt: Float[ArrayLike, ""] = 0.2,
        solver: diffrax.AbstractSolver | None = None,
        stepsize_controller: diffrax.AbstractStepSizeController | None = None,
    ):
        self.gravity = jnp.array(gravity)
        self.link_length_1 = jnp.array(link_length_1)
        self.link_length_2 = jnp.array(link_length_2)
        self.link_mass_1 = jnp.array(link_mass_1)
        self.link_mass_2 = jnp.array(link_mass_2)
        self.link_com_pos_1 = jnp.array(link_com_pos_1)
        self.link_com_pos_2 = jnp.array(link_com_pos_2)
        self.link_moi = jnp.array(link_moi)
        self.max_vel_1 = jnp.array(max_vel_1)
        self.max_vel_2 = jnp.array(max_vel_2)
        self.torque_max_noise = jnp.array(torque_max_noise)
        self.torques = jnp.array(torques)

        self.dt = jnp.array(dt)
        self.solver = solver or diffrax.Tsit5()
        is_adaptive = isinstance(self.solver, diffrax.AbstractAdaptiveSolver)
        self.dt0 = None if is_adaptive else self.dt
        if stepsize_controller is None:
            if is_adaptive:
                self.stepsize_controller = diffrax.PIDController(rtol=1e-5, atol=1e-5)
            else:
                self.stepsize_controller = diffrax.ConstantStepSize()
        else:
            self.stepsize_controller = stepsize_controller

        self.action_space = Discrete(3)
        state_high = jnp.array([1.0, 1.0, 1.0, 1.0, self.max_vel_1, self.max_vel_2])
        low = -state_high
        self.observation_space = Box(low=low, high=state_high)

    def initial(self, *, key: Key) -> AcrobotState:
        return AcrobotState(
            y=jr.uniform(key, shape=(4,), minval=-0.1, maxval=0.1), t=jnp.array(0.0)
        )

    def dynamics(
        self, t: Float[Array, ""], y: Float[Array, "4"], action: Int[Array, ""]
    ) -> Float[Array, "4"]:
        theta1, theta2, theta1_d, theta2_d = y
        a = self.torques[action]

        d1 = (
            self.link_mass_1 * self.link_com_pos_1**2
            + self.link_mass_2
            * (
                self.link_length_1**2
                + self.link_com_pos_2**2
                + 2 * self.link_length_1 * self.link_com_pos_2 * jnp.cos(theta2)
            )
            + self.link_moi
            + self.link_moi
        )
        d2 = (
            self.link_mass_2
            * (
                self.link_com_pos_2**2
                + self.link_length_1 * self.link_com_pos_2 * jnp.cos(theta2)
            )
            + self.link_moi
        )

        phi2 = (
            self.link_mass_2
            * self.link_com_pos_2
            * self.gravity
            * jnp.cos(theta1 + theta2 - jnp.pi / 2)
        )
        phi1 = (
            -self.link_mass_2
            * self.link_length_1
            * self.link_com_pos_2
            * theta2_d**2
            * jnp.sin(theta2)
            - 2
            * self.link_mass_2
            * self.link_length_1
            * self.link_com_pos_2
            * theta1_d
            * theta2_d
            * jnp.sin(theta2)
            + (
                self.link_mass_1 * self.link_com_pos_1
                + self.link_mass_2 * self.link_length_1
            )
            * self.gravity
            * jnp.cos(theta1 - jnp.pi / 2)
            + phi2
        )

        theta2_dd = (
            a
            + d2 / d1 * phi1
            - self.link_mass_2
            * self.link_length_1
            * self.link_com_pos_2
            * theta1_d**2
            * jnp.sin(theta2)
            - phi2
        ) / (self.link_mass_2 * self.link_com_pos_2**2 + self.link_moi - d2**2 / d1)
        theta1_dd = -(d2 * theta2_dd + phi1) / d1
        return jnp.array([theta1_d, theta2_d, theta1_dd, theta2_dd])

    def clip(self, y: Float[Array, "4"]) -> Float[Array, "4"]:
        joint_angle_1, joint_angle_2, joint_vel_1, joint_vel_2 = y
        joint_angle_1 = (joint_angle_1 + jnp.pi) % (2 * jnp.pi) - jnp.pi
        joint_angle_2 = (joint_angle_2 + jnp.pi) % (2 * jnp.pi) - jnp.pi
        joint_vel_1 = jnp.clip(joint_vel_1, -self.max_vel_1, self.max_vel_1)
        joint_vel_2 = jnp.clip(joint_vel_2, -self.max_vel_2, self.max_vel_2)

        return jnp.array([joint_angle_1, joint_angle_2, joint_vel_1, joint_vel_2])

    def observation(self, state: AcrobotState, *, key: Key) -> Float[Array, "4"]:
        joint_angle_1, joint_angle_2, joint_vel_1, joint_vel_2 = state.y
        return jnp.array(
            [
                jnp.cos(joint_angle_1),
                jnp.sin(joint_angle_1),
                jnp.cos(joint_angle_2),
                jnp.sin(joint_angle_2),
                joint_vel_1,
                joint_vel_2,
            ]
        )

    def reward(
        self,
        state: AcrobotState,
        action: Int[Array, ""],
        next_state: AcrobotState,
        *,
        key: Key,
    ) -> Float[Array, ""]:
        joint_angle_1, joint_angle_2 = next_state.y[0], next_state.y[1]
        done_angle = (
            -jnp.cos(joint_angle_1) - jnp.cos(joint_angle_1 + joint_angle_2) > 1.0
        )
        return done_angle.astype(float) - 1.0

    def terminal(self, state: AcrobotState, *, key: Key) -> Bool[Array, ""]:
        joint_angle_1, joint_angle_2 = state.y[0], state.y[1]
        done_angle = (
            -jnp.cos(joint_angle_1) - jnp.cos(joint_angle_1 + joint_angle_2) > 1.0
        )
        return done_angle

    def render(self, state: AcrobotState, renderer: AbstractRenderer):
        th1, th2 = state.y[0], state.y[1]

        base = jnp.array([0.0, 0.0])
        p1 = base + jnp.array(
            [self.link_length_1 * jnp.sin(th1), -self.link_length_1 * jnp.cos(th1)]
        )
        p2 = p1 + jnp.array(
            [
                self.link_length_2 * jnp.sin(th1 + th2),
                -self.link_length_2 * jnp.cos(th1 + th2),
            ]
        )

        link_w = 0.2
        joint_r = link_w / 2
        link_color = Color(0.0, 0.8, 0.8)
        joint_color = Color(0.8, 0.8, 0.0)
        goal_color = Color(0.0, 0.0, 0.0)

        renderer.clear()

        y_goal = jnp.array(1.0)
        renderer.draw_line(
            start=jnp.array([-3.0, y_goal]),
            end=jnp.array([3.0, y_goal]),
            color=goal_color,
            width=0.01,
        )

        renderer.draw_line(base, p1, color=link_color, width=link_w)
        renderer.draw_line(p1, p2, color=link_color, width=link_w)

        renderer.draw_circle(base, radius=joint_r, color=joint_color)
        renderer.draw_circle(p1, radius=joint_r, color=joint_color)

        renderer.draw()

    def default_renderer(self) -> AbstractRenderer:
        width, height = 800, 600
        transform = Transform(
            width=width,
            height=height,
            scale=140.0,
            offset=jnp.array([width / 2, height / 2]),
            y_up=True,
        )
        return PygameRenderer(
            width=width,
            height=height,
            background_color=Color(1.0, 1.0, 1.0),
            transform=transform,
        )



from __future__ import annotations

from typing import ClassVar

from jax import numpy as jnp
from jax import random as jr
from jaxtyping import Array, Float, Integer, Key, Real

from lerax.env import AbstractEnvLike, AbstractEnvLikeState
from lerax.model import MLP
from lerax.space import AbstractSpace

from .base_actor_critic import AbstractStatelessActorCriticPolicy
from .utils import ActionHead


class MLPActorCriticPolicy[
    ActType: (Float[Array, " dims"], Integer[Array, ""]),
    ObsType: Real[Array, "..."],
](AbstractStatelessActorCriticPolicy[ActType, ObsType]):
    """
    Actor–critic policy with MLP components.
    """

    name: ClassVar[str] = "MLPActorCriticPolicy"

    action_space: AbstractSpace[ActType]
    observation_space: AbstractSpace[ObsType]

    encoder: MLP
    value_head: MLP
    action_head: ActionHead

    def __init__[StateType: AbstractEnvLikeState](
        self,
        env: AbstractEnvLike[StateType, ActType, ObsType],
        *,
        feature_size: int = 16,
        feature_width: int = 64,
        feature_depth: int = 2,
        value_width: int = 64,
        value_depth: int = 2,
        action_width: int = 64,
        action_depth: int = 2,
        log_std_init: float = 0.0,
        key: Key,
    ):
        self.action_space = env.action_space
        self.observation_space = env.observation_space

        feat_key, val_key, act_key = jr.split(key, 3)

        self.encoder = MLP(
            in_size=int(jnp.array(self.observation_space.flat_size)),
            out_size=feature_size,
            width_size=feature_width,
            depth=feature_depth,
            key=feat_key,
        )

        self.value_head = MLP(
            in_size=feature_size,
            out_size="scalar",
            width_size=value_width,
            depth=value_depth,
            key=val_key,
        )

        self.action_head = ActionHead(
            self.action_space,
            feature_size,
            action_width,
            action_depth,
            key=act_key,
            log_std_init=log_std_init,
        )

    def __call__(self, observation: ObsType, *, key: Key | None = None) -> ActType:
        features = self.encoder(self.observation_space.flatten_sample(observation))
        action_dist = self.action_head(features)

        if key is None:
            action = action_dist.mode()
        else:
            action = action_dist.sample(key)

        return action

    def action_and_value(
        self, observation: ObsType, *, key: Key
    ) -> tuple[ActType, Float[Array, ""], Float[Array, ""]]:
        """
        Get an action and value from an observation.

        If `key` is provided, it will be used for sampling actions, if no key is
        provided the policy will return the most likely action.
        """
        features = self.encoder(self.observation_space.flatten_sample(observation))
        value = self.value_head(features)

        action_dist = self.action_head(features)
        action, log_prob = action_dist.sample_and_log_prob(key)

        return action, value, log_prob.sum().squeeze()

    def value(self, observation: ObsType) -> Float[Array, ""]:
        """Get the value of an observation."""
        features = self.encoder(self.observation_space.flatten_sample(observation))
        return self.value_head(features)

    def evaluate_action(
        self, observation: ObsType, action: ActType
    ) -> tuple[Float[Array, ""], Float[Array, ""], Float[Array, ""]]:
        """Evaluate an action given an observation."""
        features = self.encoder(self.observation_space.flatten_sample(observation))
        action_dist = self.action_head(features)
        value = self.value_head(features)
        log_prob = action_dist.log_prob(action)

        try:
            entropy = action_dist.entropy().squeeze()
        except NotImplementedError:
            entropy = -log_prob.mean().squeeze()  # Fallback to negative log prob mean

        return value, log_prob.sum().squeeze(), entropy



from __future__ import annotations

from typing import ClassVar, Literal

import gymnasium as gym
import jax
import numpy as np
from jax import numpy as jnp
from jax import random as jr
from jax.debug import callback as debug_callback
from jax.experimental import io_callback
from jaxtyping import Array, Bool, Float, Key

from lerax.env import AbstractEnv, AbstractEnvState
from lerax.render import AbstractRenderer
from lerax.space import AbstractSpace, Box, Dict, Discrete, Tuple


def gym_space_to_lerax_space(space: gym.Space) -> AbstractSpace:
    if isinstance(space, gym.spaces.Discrete):
        return Discrete(n=space.n)
    elif isinstance(space, gym.spaces.Box):
        return Box(low=space.low, high=space.high, shape=space.shape)
    elif isinstance(space, gym.spaces.Dict):
        return Dict({k: gym_space_to_lerax_space(s) for k, s in space.spaces.items()})
    elif isinstance(space, gym.spaces.Tuple):
        return Tuple(tuple(gym_space_to_lerax_space(s) for s in space.spaces))
    else:
        raise NotImplementedError(f"Space type {type(space)} not supported")


def lerax_to_gym_space(space: AbstractSpace) -> gym.Space:
    if isinstance(space, Discrete):
        return gym.spaces.Discrete(int(space.n), start=int(space.start))
    elif isinstance(space, Box):
        return gym.spaces.Box(
            low=np.asarray(space.low),
            high=np.asarray(space.high),
        )
    elif isinstance(space, Dict):
        return gym.spaces.Dict(
            {k: lerax_to_gym_space(s) for k, s in space.spaces.items()}
        )
    elif isinstance(space, Tuple):
        return gym.spaces.Tuple(tuple(lerax_to_gym_space(s) for s in space.spaces))
    else:
        raise NotImplementedError(f"Space type {type(space)} not supported")


def jax_to_numpy(x):
    if isinstance(x, jnp.ndarray):
        return np.asarray(x)
    return x


def to_numpy_tree(x):
    return jax.tree.map(jax_to_numpy, x)


class GymEnvState(AbstractEnvState):
    observation: Array
    reward: Float[Array, ""]
    terminal: Bool[Array, ""]
    truncated: Bool[Array, ""]


class GymToLeraxEnv(AbstractEnv[GymEnvState, Array, Array]):
    """
    Wrapper of a Gymnasium environment to make it compatible with Lerax.

    Uses jax's io_callback to wrap the env's reset and step functions.
    In general, this will be slower than a native JAX environment and prevents
    vmapped rollout. Also removes the info dict returned by Gymnasium envs since
    the shape cannot be known ahead of time.
    """

    name: ClassVar[str] = "GymnasiumEnv"

    action_space: AbstractSpace
    observation_space: AbstractSpace

    env: gym.Env

    renderer: None = None

    def __init__(self, env: gym.Env):
        self.env = env
        self.action_space = gym_space_to_lerax_space(env.action_space)
        self.observation_space = gym_space_to_lerax_space(env.observation_space)

    def _reset(self, *args, **kwargs):
        if "seed" in kwargs:
            kwargs["seed"] = int(kwargs["seed"])

        obs, _ = self.env.reset(*args, **kwargs)
        return jnp.asarray(obs)

    def initial(self, *args, key: Key, **kwargs) -> GymEnvState:
        # TODO: Determine if we want to pass a seed or not
        # I think it's a nice perk to increase reproducibility but it might
        # be unexpected for some users
        if "seed" not in kwargs:
            kwargs["seed"] = jr.randint(key, (), 0, jnp.iinfo(jnp.int32).max)

        observation = io_callback(self._reset, self.observation_space.canonical())
        return GymEnvState(
            observation=observation,
            reward=jnp.array(0.0, dtype=float),
            terminal=jnp.array(False, dtype=bool),
            truncated=jnp.array(False, dtype=bool),
        )

    def _step(self, action: Array):
        observation, reward, terminated, truncated, _ = self.env.step(
            np.asarray(action)
        )

        return (
            jnp.asarray(observation),
            jnp.asarray(reward),
            jnp.asarray(terminated),
            jnp.asarray(truncated),
        )

    def transition(self, state: GymEnvState, action: Array, *, key: Key) -> GymEnvState:
        observation, reward, terminated, truncated = io_callback(
            self._step,
            (
                self.observation_space.canonical(),
                jnp.array(0.0, dtype=float),
                jnp.array(False, dtype=bool),
                jnp.array(False, dtype=bool),
            ),
            action,
        )

        return GymEnvState(
            observation=observation,
            reward=reward,
            terminal=terminated,
            truncated=truncated,
        )

    def observation(self, state: GymEnvState, *, key: Key) -> Array:
        return state.observation

    def reward(
        self, state: GymEnvState, action: Array, next_state: GymEnvState, *, key: Key
    ) -> Float[Array, ""]:
        return next_state.reward

    def terminal(self, state: GymEnvState, *, key: Key) -> Bool[Array, ""]:
        return state.terminal

    def truncate(self, state: GymEnvState) -> Bool[Array, ""]:
        return state.truncated

    def state_info(self, state: GymEnvState) -> dict:
        return {}

    def transition_info(
        self, state: GymEnvState, action: Array, next_state: GymEnvState
    ) -> dict:
        return {}

    def render(self, state: GymEnvState, renderer: AbstractRenderer):
        raise NotImplementedError("Rendering not implemented for GymToLeraxEnv")

    def default_renderer(self) -> AbstractRenderer:
        raise NotImplementedError("Default renderer not implemented for GymToLeraxEnv.")

    def close(self):
        debug_callback(self.env.close, ordered=True)


class LeraxToGymEnv[StateType: AbstractEnvState](gym.Env):
    """
    Wrapper of an Lerax environment to make it compatible with Gymnasium.

    Executes the Lerax env directly (Python side). Keeps an internal eqx state and PRNG.
    """

    metadata: dict = {"render_modes": ["human"]}

    action_space: gym.Space
    observation_space: gym.Space

    render_mode: str | None = None

    env: AbstractEnv[StateType, Array, Array]
    state: StateType
    key: Key

    def __init__(
        self,
        env: AbstractEnv[StateType, Array, Array],
        render_mode: Literal["human"] | None = None,
    ):
        self.key = jr.key(0)

        self.env = env

        self.action_space = lerax_to_gym_space(env.action_space)
        self.observation_space = lerax_to_gym_space(env.observation_space)

        self.render_mode = render_mode
        # TODO: Actually handle rendering

    def reset(self, *, seed: int | None = None, options: dict | None = None):
        if seed is not None:
            self.key = jr.key(int(seed))

        self.key, reset_key = jr.split(self.key)
        self.state, obs, info = self.env.reset(key=reset_key)
        return jax_to_numpy(obs), to_numpy_tree(info)

    def step(self, action):
        self.key, step_key = jr.split(self.key)
        self.state, obs, rew, term, trunc, info = self.env.step(
            self.state, jnp.asarray(action), key=step_key
        )

        return (
            jax_to_numpy(obs),
            float(jnp.asarray(rew)),
            bool(jnp.asarray(term)),
            bool(jnp.asarray(trunc)),
            to_numpy_tree(info),
        )

    def render(self):
        raise NotImplementedError("Rendering not implemented for LeraxToGymEnv")

    def close(self): ...



from __future__ import annotations

from abc import abstractmethod

import equinox as eqx
import jax
from jax import numpy as jnp
from jax import random as jr
from jaxtyping import Key


class AbstractBuffer(eqx.Module):
    """Base class for buffers."""

    def batches[SelfType: AbstractBuffer](
        self: SelfType,
        batch_size: int,
        *,
        key: Key | None = None,
        batch_axes: tuple[int, ...] | int | None = None,
    ) -> SelfType:
        ndim = len(self.shape)

        if batch_axes is None:
            axes = tuple(range(ndim))
        elif isinstance(batch_axes, int):
            axes = (batch_axes,)
        else:
            axes = tuple(batch_axes)

        axes = tuple(a + ndim if a < 0 else a for a in axes)
        if len(set(axes)) != len(axes) or any(a < 0 or a >= ndim for a in axes):
            raise ValueError(f"Invalid batch_axes {batch_axes} for array ndim={ndim}.")

        def flatten_selected(x):
            moved = jnp.moveaxis(x, axes, tuple(range(len(axes))))
            lead = 1
            for i in range(len(axes)):
                lead *= moved.shape[i]

            return moved.reshape((lead,) + moved.shape[len(axes) :])

        flat_self = jax.tree.map(flatten_selected, self)

        total = flat_self.rewards.shape[0]
        indices = jnp.arange(total) if key is None else jr.permutation(key, total)

        if total % batch_size != 0:
            total_trim = total - (total % batch_size)
            indices = indices[:total_trim]

        indices = indices.reshape(-1, batch_size)

        return jax.tree.map(
            lambda x: jnp.take(x, indices, axis=0) if isinstance(x, jnp.ndarray) else x,
            flat_self,
        )

    @property
    @abstractmethod
    def shape(self) -> tuple[int, ...]:
        """Return the shape of the buffer."""



from __future__ import annotations

from abc import abstractmethod
from datetime import datetime

import equinox as eqx
import jax
import optax
from jax import numpy as jnp
from jax import random as jr
from jaxtyping import Key

from lerax.env import AbstractEnvLike
from lerax.policy import (
    AbstractPolicy,
    AbstractStatefulPolicy,
    AbstractStatelessPolicy,
)
from lerax.utils import filter_scan

from .utils import IterationCarry, JITProgressBar, JITSummaryWriter, StepCarry


class AbstractAlgorithm[PolicyType: AbstractPolicy](eqx.Module):
    """Base class for RL algorithms."""

    optimizer: eqx.AbstractVar[optax.GradientTransformation]

    num_envs: eqx.AbstractVar[int]
    num_steps: eqx.AbstractVar[int]

    def init_iteration_carry[WrappedPolicyType: AbstractStatefulPolicy](
        self,
        env: AbstractEnvLike,
        policy: WrappedPolicyType,
        *,
        key: Key,
    ) -> IterationCarry[WrappedPolicyType]:
        if self.num_envs == 1:
            step_carry = StepCarry.initial(env, policy, key)
        else:
            step_carry = jax.vmap(StepCarry.initial, in_axes=(None, None, 0))(
                env, policy, jr.split(key, self.num_envs)
            )

        return IterationCarry(
            jnp.array(0, dtype=int),
            step_carry,
            policy,
            self.optimizer.init(eqx.filter(policy, eqx.is_inexact_array)),
        )

    def init_tensorboard(
        self,
        env: AbstractEnvLike,
        policy: AbstractPolicy,
        tb_log: str | bool,
    ) -> JITSummaryWriter | None:
        if tb_log is False:
            return None

        if tb_log is True:
            tb_log = f"logs/{policy.name}_{env.name}_{datetime.now().strftime('%Y%m%d_%H%M%S')}"

        return JITSummaryWriter(tb_log)

    def init_progress_bar(
        self,
        env: AbstractEnvLike,
        policy: AbstractPolicy,
        total_timesteps: int,
        show_progress_bar: bool,
    ) -> JITProgressBar | None:
        if show_progress_bar:
            name = f"Training {policy.name} on {env.name}"
            progress_bar = JITProgressBar(name, total=total_timesteps)
            progress_bar.start()
            return progress_bar
        else:
            return None

    @abstractmethod
    def iteration[WrappedPolicyType: AbstractStatefulPolicy](
        self,
        env: AbstractEnvLike,
        carry: IterationCarry[WrappedPolicyType],
        *,
        key: Key,
        progress_bar: JITProgressBar | None,
        tb_writer: JITSummaryWriter | None,
    ) -> IterationCarry[WrappedPolicyType]:
        """Perform a single iteration of training."""

    # TODO: Add support for callbacks
    def learn(
        self,
        env: AbstractEnvLike,
        policy: PolicyType,
        total_timesteps: int,
        *,
        key: Key,
        show_progress_bar: bool = False,
        tb_log: str | bool = False,
    ) -> PolicyType:
        if not isinstance(policy, AbstractStatefulPolicy):
            wrapped_policy = policy.into_stateful()
        else:
            wrapped_policy = policy

        init_key, learn_key = jr.split(key, 2)

        carry = self.init_iteration_carry(env, wrapped_policy, key=init_key)

        progress_bar = self.init_progress_bar(
            env, wrapped_policy, total_timesteps, show_progress_bar
        )
        tb_writer = self.init_tensorboard(env, wrapped_policy, tb_log)
        num_iterations = total_timesteps // (self.num_steps * self.num_envs)

        @eqx.filter_jit
        def learn(carry: IterationCarry) -> IterationCarry:
            def scan_iteration(carry: tuple[IterationCarry, Key], _):
                it_carry, key = carry
                iter_key, next_key = jr.split(key, 2)
                it_carry = self.iteration(
                    env,
                    it_carry,
                    key=iter_key,
                    progress_bar=progress_bar,
                    tb_writer=tb_writer,
                )
                return (it_carry, next_key), None

            (carry, _), _ = filter_scan(
                scan_iteration, (carry, learn_key), length=num_iterations
            )

            return carry

        carry = learn(carry)

        if progress_bar is not None:
            progress_bar.stop()

        if isinstance(policy, AbstractStatelessPolicy):
            return carry.policy.policy
        else:
            return carry.policy



from __future__ import annotations

from abc import abstractmethod

import equinox as eqx
from jax import numpy as jnp
from jax import random as jr
from jaxtyping import Array, Float, Int, Key

from lerax.distribution import (
    AbstractDistribution,
    Bernoulli,
    Categorical,
    SquashedMultivariateNormalDiag,
    SquashedNormal,
)
from lerax.model import AbstractModel
from lerax.space import AbstractSpace, Box, Discrete, MultiBinary


class AbstractActionLayer[ActType](AbstractModel):
    """Layer that produces action distributions given inputs."""

    mapping: eqx.AbstractVar[eqx.nn.Linear]

    @abstractmethod
    def __call__(
        self,
        inputs: Float[Array, " latent_dim"],
    ) -> AbstractDistribution[ActType]:
        """Produce an action distribution given inputs."""


class BoxActionLayer(AbstractActionLayer[Float[Array, " action_dim"]]):

    scalar: bool
    mapping: eqx.nn.Linear
    log_std: Float[Array, " action_dim"]

    def __init__(
        self, latent_dim: int, action_space: Box, *, key: Key, log_std_init: float = 0.0
    ):
        if action_space.shape:
            self.scalar = False
            self.mapping = eqx.nn.Linear(
                latent_dim, int(action_space.flat_size), key=key
            )
            self.log_std = jnp.full(int(action_space.flat_size), log_std_init)
        else:
            self.scalar = True
            self.mapping = eqx.nn.Linear(latent_dim, "scalar", key=key)
            self.log_std = jnp.array(log_std_init)

    def __call__(
        self, inputs: Float[Array, " latent_dim"]
    ) -> SquashedNormal | SquashedMultivariateNormalDiag:
        if self.scalar:
            return SquashedNormal(loc=self.mapping(inputs), scale=jnp.exp(self.log_std))
        else:
            return SquashedMultivariateNormalDiag(
                loc=self.mapping(inputs), scale_diag=jnp.exp(self.log_std)
            )


class CategoricalActionLayer(AbstractActionLayer[Int[Array, ""]]):

    mapping: eqx.nn.Linear

    def __init__(self, latent_dim: int, action_space: Discrete, *, key: Key):
        self.mapping = eqx.nn.Linear(latent_dim, int(action_space.n), key=key)

    def __call__(self, inputs: Float[Array, " latent_dim"]) -> Categorical:
        return Categorical(logits=self.mapping(inputs))


class BernoulliActionLayer(AbstractActionLayer[Int[Array, ""]]):

    mapping: eqx.nn.Linear

    def __init__(self, latent_dim: int, action_space: MultiBinary, *, key: Key):
        self.mapping = eqx.nn.Linear(latent_dim, int(action_space.n), key=key)

    def __call__(self, inputs: Float[Array, " latent_dim"]) -> Bernoulli:
        return Bernoulli(logits=self.mapping(inputs))


def make_action_layer(
    latent_dim: int, action_space: AbstractSpace, *, key: Key, log_std_init: float = 0.0
) -> AbstractActionLayer:
    """Create an action layer based on the action space."""

    if isinstance(action_space, Box):
        return BoxActionLayer(
            latent_dim, action_space, key=key, log_std_init=log_std_init
        )
    elif isinstance(action_space, Discrete):
        return CategoricalActionLayer(latent_dim, action_space, key=key)
    elif isinstance(action_space, MultiBinary):
        return BernoulliActionLayer(latent_dim, action_space, key=key)
    else:
        raise NotImplementedError(f"Action space {type(action_space)} not supported.")


class ActionHead[ActType](AbstractModel):
    mlp: eqx.nn.MLP | None
    action_layer: AbstractActionLayer[ActType]

    def __init__(
        self,
        action_space: AbstractSpace,
        latent_dim: int,
        width_size: int,
        depth: int,
        *,
        key: Key,
        log_std_init: float = 0.0,
    ):
        mlp_key, action_key = jr.split(key, 2)

        depth = max(0, depth - 1)
        if depth < 0:
            self.mlp = None
        else:
            self.mlp = eqx.nn.MLP(
                in_size=latent_dim,
                out_size=latent_dim,
                width_size=width_size,
                depth=depth,
                key=mlp_key,
            )
        self.action_layer = make_action_layer(
            latent_dim, action_space, key=action_key, log_std_init=log_std_init
        )

    def __call__(
        self, features: Float[Array, " latent_dim"]
    ) -> AbstractDistribution[ActType]:
        if self.mlp is not None:
            features = self.mlp(features)
        return self.action_layer(features)



from __future__ import annotations

from typing import ClassVar

import diffrax
from jax import numpy as jnp
from jax import random as jr
from jaxtyping import Array, ArrayLike, Bool, Float, Int, Key

from lerax.render import AbstractRenderer, Color, PygameRenderer, Transform
from lerax.space import Box, Discrete

from .base_classic_control import (
    AbstractClassicControlEnv,
    AbstractClassicControlEnvState,
)


class CartPoleState(AbstractClassicControlEnvState):
    y: Float[Array, "4"]
    t: Float[Array, ""]


class CartPole(
    AbstractClassicControlEnv[CartPoleState, Int[Array, ""], Float[Array, "4"]]
):
    name: ClassVar[str] = "CartPole"

    action_space: Discrete
    observation_space: Box

    gravity: Float[Array, ""]
    cart_mass: Float[Array, ""]
    pole_mass: Float[Array, ""]
    total_mass: Float[Array, ""]
    length: Float[Array, ""]
    polemass_length: Float[Array, ""]
    force_mag: Float[Array, ""]
    theta_threshold_radians: Float[Array, ""]
    x_threshold: Float[Array, ""]

    dt: Float[Array, ""]
    solver: diffrax.AbstractSolver
    dt0: Float[Array, ""] | None
    stepsize_controller: diffrax.AbstractStepSizeController

    def __init__(
        self,
        *,
        gravity: Float[ArrayLike, ""] = 9.8,
        cart_mass: Float[ArrayLike, ""] = 1.0,
        pole_mass: Float[ArrayLike, ""] = 0.1,
        half_length: Float[ArrayLike, ""] = 0.5,
        force_mag: Float[ArrayLike, ""] = 10.0,
        theta_threshold_radians: Float[ArrayLike, ""] = 12 * 2 * jnp.pi / 360,
        x_threshold: Float[ArrayLike, ""] = 2.4,
        dt: Float[ArrayLike, ""] = 0.02,
        solver: diffrax.AbstractSolver | None = None,
        stepsize_controller: diffrax.AbstractStepSizeController | None = None,
    ):
        self.gravity = jnp.array(gravity)
        self.cart_mass = jnp.array(cart_mass)
        self.pole_mass = jnp.array(pole_mass)
        self.total_mass = self.pole_mass + self.cart_mass
        self.length = jnp.array(half_length)
        self.polemass_length = self.pole_mass * self.length
        self.force_mag = jnp.array(force_mag)

        self.dt = jnp.array(dt)
        self.solver = solver or diffrax.Tsit5()
        is_adaptive = isinstance(self.solver, diffrax.AbstractAdaptiveSolver)
        self.dt0 = None if is_adaptive else self.dt
        if stepsize_controller is None:
            if is_adaptive:
                self.stepsize_controller = diffrax.PIDController(rtol=1e-5, atol=1e-5)
            else:
                self.stepsize_controller = diffrax.ConstantStepSize()
        else:
            self.stepsize_controller = stepsize_controller

        self.theta_threshold_radians = jnp.array(theta_threshold_radians)
        self.x_threshold = jnp.array(x_threshold)

        self.action_space = Discrete(2)
        high = jnp.array(
            [
                self.x_threshold * 2,
                jnp.inf,
                self.theta_threshold_radians * 2,
                jnp.inf,
            ],
        )
        self.observation_space = Box(-high, high)

    def initial(self, *, key: Key) -> CartPoleState:
        return CartPoleState(
            y=jr.uniform(key, (4,), minval=-0.05, maxval=0.05), t=jnp.array(0.0)
        )

    def dynamics(
        self, t: Float[Array, ""], y: Float[Array, "4"], action: Int[Array, ""]
    ) -> Float[Array, "4"]:
        _, x_dot, theta, theta_dot = y
        force = (action * 2 - 1) * self.force_mag

        temp = (
            force + self.polemass_length * theta_dot**2 * jnp.sin(theta)
        ) / self.total_mass
        theta_dd = (self.gravity * jnp.sin(theta) - jnp.cos(theta) * temp) / (
            self.length
            * (4.0 / 3.0 - self.pole_mass * (jnp.cos(theta) ** 2) / self.total_mass)
        )
        x_dd = temp - self.polemass_length * theta_dd * jnp.cos(theta) / self.total_mass

        return jnp.array([x_dot, x_dd, theta_dot, theta_dd])

    def clip(self, y: Float[Array, "4"]) -> Float[Array, "4"]:
        return y

    def observation(self, state: CartPoleState, *, key: Key) -> Float[Array, "4"]:
        return state.y

    def reward(
        self,
        state: CartPoleState,
        action: Int[Array, ""],
        next_state: CartPoleState,
        *,
        key: Key,
    ) -> Float[Array, ""]:
        return jnp.array(1.0)

    def terminal(self, state: CartPoleState, *, key: Key) -> Bool[Array, ""]:
        x, theta = state.y[0], state.y[2]
        within_x = (x >= -self.x_threshold) & (x <= self.x_threshold)
        within_theta = (theta >= -self.theta_threshold_radians) & (
            theta <= self.theta_threshold_radians
        )
        return ~(within_x & within_theta)

    def render(self, state: CartPoleState, renderer: AbstractRenderer):
        x, theta = state.y[0], state.y[2]

        renderer.clear()

        # Ground
        renderer.draw_line(
            start=jnp.array((-10.0, 0.0)),
            end=jnp.array((10.0, 0.0)),
            color=Color(0.0, 0.0, 0.0),
            width=0.01,
        )
        # Cart
        cart_w, cart_h = 0.3, 0.15
        cart_col = Color(0.0, 0.0, 0.0)
        renderer.draw_rect(jnp.array((x, 0.0)), w=cart_w, h=cart_h, color=cart_col)
        # Pole
        pole_start = jnp.asarray((x, cart_h / 4))
        pole_end = pole_start + self.length * jnp.asarray(
            [jnp.sin(theta), jnp.cos(theta)]
        )
        pole_col = Color(0.8, 0.6, 0.4)
        renderer.draw_line(pole_start, pole_end, color=pole_col, width=0.05)
        # Pole Hinge
        hinge_r = 0.025
        hinge_col = Color(0.5, 0.5, 0.5)
        renderer.draw_circle(pole_start, radius=hinge_r, color=hinge_col)

        renderer.draw()

    def default_renderer(self) -> AbstractRenderer:
        width, height = 800, 450
        transform = Transform(
            scale=200.0,
            offset=jnp.array([width / 2, height * 0.1]),
            width=width,
            height=height,
            y_up=True,
        )
        return PygameRenderer(
            width=width,
            height=height,
            background_color=Color(1.0, 1.0, 1.0),
            transform=transform,
        )



from __future__ import annotations

from typing import ClassVar

import diffrax
import equinox as eqx
from jax import numpy as jnp
from jax import random as jr
from jaxtyping import Array, Float, Integer, Key, Real

from lerax.env import AbstractEnvLike, AbstractEnvLikeState
from lerax.model import MLP, MLPNeuralCDE, NCDEState
from lerax.space import AbstractSpace

from .base_actor_critic import (
    AbstractPolicyState,
    AbstractStatefulActorCriticPolicy,
)
from .utils import ActionHead


class NCDEPolicyState(AbstractPolicyState):
    t: Float[Array, ""]
    cde: NCDEState

    def __init__(self, *, t: Float[Array, ""] = jnp.array(0.0), cde: NCDEState):
        self.t = jnp.asarray(t, dtype=float)
        self.cde = cde


class NCDEActorCriticPolicy[
    ActType: (Float[Array, " dims"], Integer[Array, ""]),
    ObsType: Real[Array, "..."],
](AbstractStatefulActorCriticPolicy[NCDEPolicyState, ActType, ObsType]):
    """
    Actor–critic with a shared MLPNeuralCDE encoder and MLP heads.
    """

    name: ClassVar[str] = "NCDEActorCriticPolicy"

    action_space: AbstractSpace[ActType]
    observation_space: AbstractSpace[ObsType]

    encoder: MLPNeuralCDE
    value_head: MLP
    action_head: ActionHead

    dt: float = eqx.field(static=True)

    def __init__[StateType: AbstractEnvLikeState](
        self,
        env: AbstractEnvLike[StateType, ActType, ObsType],
        *,
        solver: diffrax.AbstractSolver | None = None,
        feature_size: int = 4,
        latent_size: int = 4,
        field_width: int = 8,
        field_depth: int = 1,
        initial_width: int = 16,
        initial_depth: int = 1,
        value_width: int = 16,
        value_depth: int = 1,
        action_width: int = 16,
        action_depth: int = 1,
        history_length: int = 4,
        dt: float = 1.0,
        log_std_init: float = 0.0,
        key: Key,
    ):
        self.action_space = env.action_space
        self.observation_space = env.observation_space
        self.dt = float(dt)

        enc_key, val_key, act_key = jr.split(key, 3)

        self.encoder = MLPNeuralCDE(
            in_size=int(jnp.array(self.observation_space.flat_size)),
            latent_size=latent_size,
            solver=solver,
            field_width=field_width,
            field_depth=field_depth,
            initial_width=initial_width,
            initial_depth=initial_depth,
            time_in_input=False,
            history_length=history_length,
            key=enc_key,
        )

        self.value_head = MLP(
            in_size=latent_size,
            out_size="scalar",
            width_size=value_width,
            depth=value_depth,
            key=val_key,
        )

        self.action_head = ActionHead(
            self.action_space,
            feature_size,
            action_width,
            action_depth,
            key=act_key,
            log_std_init=log_std_init,
        )

    def _step_encoder(
        self, state: NCDEPolicyState, obs: ObsType
    ) -> tuple[NCDEPolicyState, Float[Array, " feat"]]:
        t_next = state.t + self.dt
        cde_state, y = self.encoder(
            state.cde, t_next, self.observation_space.flatten_sample(obs)
        )
        return NCDEPolicyState(t=t_next, cde=cde_state), y

    def reset(self) -> NCDEPolicyState:
        return NCDEPolicyState(t=jnp.array(0.0), cde=self.encoder.reset())

    def __call__(
        self, state: NCDEPolicyState, observation: ObsType, *, key: Key | None = None
    ) -> tuple[NCDEPolicyState, ActType]:
        state, features = self._step_encoder(state, observation)
        action_dist = self.action_head(features)

        if key is None:
            action = action_dist.mode()
        else:
            action = action_dist.sample(key)

        return state, action

    def action_and_value(
        self, state: NCDEPolicyState, observation: ObsType, *, key: Key
    ) -> tuple[NCDEPolicyState, ActType, Float[Array, ""], Float[Array, ""]]:
        state, features = self._step_encoder(state, observation)

        value = self.value_head(features)

        dist = self.action_head(features)
        action, log_prob = dist.sample_and_log_prob(key)

        return state, action, value, log_prob.sum().squeeze()

    def evaluate_action(
        self, state: NCDEPolicyState, observation: ObsType, action: ActType
    ) -> tuple[NCDEPolicyState, Float[Array, ""], Float[Array, ""], Float[Array, ""]]:
        state, features = self._step_encoder(state, observation)
        dist = self.action_head(features)
        value = self.value_head(features)
        log_prob = dist.log_prob(action)

        try:
            entropy = dist.entropy().squeeze()
        except NotImplementedError:
            entropy = -log_prob.mean().squeeze()

        return state, value, log_prob.sum().squeeze(), entropy

    def value(
        self, state: NCDEPolicyState, observation: ObsType
    ) -> tuple[NCDEPolicyState, Float[Array, ""]]:
        state, feats = self._step_encoder(state, observation)
        return state, self.value_head(feats)



from __future__ import annotations

from abc import abstractmethod

import equinox as eqx
from jaxtyping import Array, Float, Key

from lerax.space import AbstractSpace

from ..base_policy import (
    AbstractPolicy,
    AbstractPolicyState,
    AbstractStatefulPolicy,
    AbstractStatefulWrapper,
    AbstractStatelessPolicy,
    NullStatefulPolicyState,
)


class AbstractActorCriticPolicy[ActType, ObsType](AbstractPolicy[ActType, ObsType]):
    """Base class for actor-critic policies."""

    name: eqx.AbstractClassVar[str]
    action_space: eqx.AbstractVar[AbstractSpace[ActType]]
    observation_space: eqx.AbstractVar[AbstractSpace[ObsType]]


class AbstractStatelessActorCriticPolicy[
    ActType,
    ObsType,
](
    AbstractActorCriticPolicy[ActType, ObsType],
    AbstractStatelessPolicy[ActType, ObsType],
):
    """
    Base class for stateless actor-critic policies.

    This class is intended for policies that do not maintain an internal state.
    Stateless policies can be converted to stateful ones using into_stateful().
    """

    name: eqx.AbstractClassVar[str]
    action_space: eqx.AbstractVar[AbstractSpace[ActType]]
    observation_space: eqx.AbstractVar[AbstractSpace[ObsType]]

    @abstractmethod
    def __call__(self, observation: ObsType, *, key: Key | None = None) -> ActType:
        """
        Get an action from an observation.

        If `key` is provided, it will be used for sampling actions, if no key is
        provided the policy will return the most likely action.
        """

    @abstractmethod
    def action_and_value(
        self, observation: ObsType, *, key: Key
    ) -> tuple[ActType, Float[Array, ""], Float[Array, ""]]:
        """
        Get an action and value from an observation.

        If `key` is provided, it will be used for sampling actions, if no key is
        provided the policy will return the most likely action.
        """

    @abstractmethod
    def evaluate_action(
        self, observation: ObsType, action: ActType
    ) -> tuple[Float[Array, ""], Float[Array, ""], Float[Array, ""]]:
        """Evaluate an action given an observation."""

    @abstractmethod
    def value(self, observation: ObsType) -> Float[Array, ""]:
        """Get the value of an observation."""

    def into_stateful[SelfType: AbstractStatelessActorCriticPolicy](
        self: SelfType,
    ) -> ActorCriticStatefulWrapper[SelfType, ActType, ObsType]:
        return ActorCriticStatefulWrapper(self)


class AbstractStatefulActorCriticPolicy[
    StateType: AbstractPolicyState, ActType, ObsType
](
    AbstractActorCriticPolicy[ActType, ObsType],
    AbstractStatefulPolicy[StateType, ActType, ObsType],
):
    """
    Base class for stateful actor-critic policies.

    This class is intended for policies that maintain an internal state, such as RNNs.
    """

    name: eqx.AbstractClassVar[str]
    action_space: eqx.AbstractVar[AbstractSpace[ActType]]
    observation_space: eqx.AbstractVar[AbstractSpace[ObsType]]

    @abstractmethod
    def __call__(
        self, state: StateType, observation: ObsType, *, key: Key | None = None
    ) -> tuple[StateType, ActType]:
        """
        Get an action from an observation.

        If `key` is provided, it will be used for sampling actions, if no key is
        provided the policy will return the most likely action.
        """

    @abstractmethod
    def action_and_value(
        self, state: StateType, observation: ObsType, *, key: Key
    ) -> tuple[StateType, ActType, Float[Array, ""], Float[Array, ""]]:
        """
        Get an action and value from an observation.

        If `key` is provided, it will be used for sampling actions, if no key is
        provided the policy will return the most likely action.
        """

    @abstractmethod
    def evaluate_action(
        self, state: StateType, observation: ObsType, action: ActType
    ) -> tuple[StateType, Float[Array, ""], Float[Array, ""], Float[Array, ""]]:
        """Evaluate an action given an observation."""

    @abstractmethod
    def value(
        self, state: StateType, observation: ObsType
    ) -> tuple[StateType, Float[Array, ""]]:
        """Get the value of an observation."""


class ActorCriticStatefulWrapper[
    PolicyType: AbstractStatelessActorCriticPolicy,
    ActType,
    ObsType,
](
    AbstractStatefulActorCriticPolicy[NullStatefulPolicyState, ActType, ObsType],
    AbstractStatefulWrapper[PolicyType, ActType, ObsType],
):
    policy: PolicyType

    def __init__(self, policy: PolicyType):
        self.policy = policy

    def __call__(
        self,
        state: NullStatefulPolicyState,
        observation: ObsType,
        *,
        key: Key | None = None,
    ) -> tuple[NullStatefulPolicyState, ActType]:
        return state, self.policy(observation, key=key)

    def action_and_value(
        self, state: NullStatefulPolicyState, observation: ObsType, *, key: Key
    ) -> tuple[NullStatefulPolicyState, ActType, Float[Array, ""], Float[Array, ""]]:
        return state, *self.policy.action_and_value(observation, key=key)

    def evaluate_action(
        self, state: NullStatefulPolicyState, observation: ObsType, action: ActType
    ) -> tuple[
        NullStatefulPolicyState, Float[Array, ""], Float[Array, ""], Float[Array, ""]
    ]:
        return state, *self.policy.evaluate_action(observation, action)

    def value(
        self, state: NullStatefulPolicyState, observation: ObsType
    ) -> tuple[NullStatefulPolicyState, Float[Array, ""]]:
        return state, self.policy.value(observation)



from __future__ import annotations

from abc import abstractmethod

import equinox as eqx
import optax
from jax import lax
from jax import numpy as jnp
from jax import random as jr
from jaxtyping import Key, Scalar

from lerax.buffer import ReplayBuffer
from lerax.env import AbstractEnvLike
from lerax.policy import AbstractPolicy, AbstractStatefulPolicy
from lerax.utils import filter_scan

from .base_algorithm import AbstractAlgorithm
from .utils import (
    EpisodeStats,
    IterationCarry,
    JITProgressBar,
    JITSummaryWriter,
    StepCarry,
)


class AbstractOffPolicyAlgorithm[PolicyType: AbstractPolicy](
    AbstractAlgorithm[PolicyType]
):
    buffer_size: eqx.AbstractVar[int]
    optimizer: eqx.AbstractVar[optax.GradientTransformation]

    gamma: eqx.AbstractVar[float]

    num_envs: eqx.AbstractVar[int]
    num_steps: eqx.AbstractVar[int]
    batch_size: eqx.AbstractVar[int]

    def step(
        self,
        env: AbstractEnvLike,
        policy: AbstractStatefulPolicy,
        carry: StepCarry,
        key: Key,
    ) -> tuple[StepCarry, ReplayBuffer]:
        (
            action_key,
            transition_key,
            observation_key,
            reward_key,
            terminal_key,
            next_observation_key,
            reset_key,
        ) = jr.split(key, 7)

        observation = env.observation(carry.env_state, key=observation_key)
        next_policy_state, action = policy(
            carry.policy_state, observation, key=action_key
        )

        next_env_state = env.transition(carry.env_state, action, key=transition_key)

        reward = env.reward(carry.env_state, action, next_env_state, key=reward_key)
        termination = env.terminal(carry.env_state, key=terminal_key)
        truncation = env.truncate(carry.env_state)
        done = termination | truncation
        timeout = truncation & ~termination

        next_episode_stats = carry.episode_stats.next(reward, done)

        next_observation = env.observation(next_env_state, key=next_observation_key)

        next_env_state = lax.cond(
            done, lambda: env.initial(key=reset_key), lambda: next_env_state
        )

        next_policy_state = lax.cond(
            done, lambda: policy.reset(), lambda: next_policy_state
        )

        return StepCarry(
            next_env_state, next_policy_state, next_episode_stats
        ), ReplayBuffer(
            observations=observation,
            next_observations=next_observation,
            actions=action,
            rewards=reward,
            dones=done,
            timeouts=timeout,
            states=carry.policy_state,
        )

    def collect_rollout(
        self,
        env: AbstractEnvLike,
        policy: AbstractStatefulPolicy,
        carry: StepCarry,
        key: Key,
    ) -> tuple[StepCarry, ReplayBuffer, EpisodeStats]:
        def scan_step(
            carry: StepCarry, key: Key
        ) -> tuple[StepCarry, tuple[ReplayBuffer, EpisodeStats]]:
            carry, buffer = self.step(env, policy, carry, key)
            return carry, (buffer, carry.episode_stats)

        carry, (replay_buffer, episode_stats) = filter_scan(
            scan_step, carry, jr.split(key, self.num_steps)
        )

        return carry, replay_buffer, episode_stats

    @abstractmethod
    def train[WrapperPolicyType: AbstractStatefulPolicy](
        self,
        policy: WrapperPolicyType,
        opt_state: optax.OptState,
        buffer: ReplayBuffer,
        *,
        key: Key,
    ) -> tuple[WrapperPolicyType, optax.OptState, dict[str, Scalar]]:
        """Trains the policy using data from the replay buffer."""

    def iteration[WrapperPolicyType: AbstractStatefulPolicy](
        self,
        env: AbstractEnvLike,
        carry: IterationCarry[WrapperPolicyType],
        *,
        key: Key,
        progress_bar: JITProgressBar | None,
        tb_writer: JITSummaryWriter | None,
    ) -> IterationCarry[WrapperPolicyType]:
        rollout_key, train_key = jr.split(key, 2)
        if self.num_steps == 1:
            step_carry, replay_buffer, episode_stats = self.collect_rollout(
                env, carry.policy, carry.step_carry, key=rollout_key
            )
        else:
            step_carry, replay_buffer, episode_stats = eqx.filter_vmap(
                self.collect_rollout, in_axes=(None, None, 0, 0)
            )(
                env,
                carry.policy,
                carry.step_carry,
                jr.split(rollout_key, self.num_envs),
            )

        policy, opt_state, log = self.train(
            carry.policy, carry.opt_state, replay_buffer, key=train_key
        )

        if progress_bar is not None:
            progress_bar.update(self.num_envs * self.num_steps)

        if tb_writer is not None:
            first_step = carry.iteration_count * self.num_steps * self.num_envs
            final_step = first_step + self.num_steps * self.num_envs - 1
            log["learning_rate"] = optax.tree_utils.tree_get(
                opt_state, "learning_rate", jnp.nan
            )
            tb_writer.add_dict(log, prefix="train", global_step=final_step)
            tb_writer.log_episode_stats(episode_stats, first_step=first_step)

        return IterationCarry(carry.iteration_count + 1, step_carry, policy, opt_state)

    def learn(
        self,
        env: AbstractEnvLike,
        policy: PolicyType,
        total_timesteps: int,
        *,
        key: Key,
        show_progress_bar: bool = False,
        tb_log: str | bool = False,
    ) -> PolicyType:
        raise NotImplementedError



from __future__ import annotations

from jax import numpy as jnp
from jaxtyping import (
    Array,
    ArrayLike,
    Bool,
    Float,
    PyTree,
)

from lerax.policy import AbstractPolicyState

from .base_buffer import AbstractBuffer


class ReplayBuffer[StateType: AbstractPolicyState, ActType, ObsType](AbstractBuffer):
    """
    ReplayBuffer used by off-policy algorithms.

    The buffer is implemented as a fixed-size circular buffer. New transitions
    overwrite the oldest ones once the capacity is reached.

    Observations and actions are stored as PyTrees of arrays, with a leading
    buffer dimension. This supports arbitrary nested structures as long as
    they are JAX arrays.
    """

    observations: PyTree[ObsType]
    next_observations: PyTree[ObsType]
    actions: PyTree[ActType]
    rewards: Float[Array, " capacity"]
    dones: Bool[Array, " capacity"]
    timeouts: Bool[Array, " capacity"]
    states: StateType

    def __init__(
        self,
        observations: PyTree[ObsType],
        next_observations: PyTree[ObsType],
        actions: PyTree[ActType],
        rewards: Float[ArrayLike, " capacity"],
        dones: Bool[ArrayLike, " capacity"],
        timeouts: Bool[ArrayLike, " capacity"],
        states: StateType,
    ):
        self.observations = observations
        self.next_observations = next_observations
        self.actions = actions
        self.rewards = jnp.asarray(rewards, dtype=float)
        self.dones = jnp.asarray(dones, dtype=bool)
        self.timeouts = jnp.asarray(timeouts, dtype=bool)
        self.states = states

    @property
    def shape(self) -> tuple[int, ...]:
        return self.rewards.shape



from .actor_critic import (
    AbstractActorCriticPolicy,
    AbstractStatefulActorCriticPolicy,
    AbstractStatelessActorCriticPolicy,
    MLPActorCriticPolicy,
    NCDEActorCriticPolicy,
)
from .base_policy import (
    AbstractPolicy,
    AbstractPolicyState,
    AbstractStatefulPolicy,
    AbstractStatelessPolicy,
)

__all__ = [
    "AbstractPolicy",
    "AbstractStatefulPolicy",
    "AbstractStatelessPolicy",
    "AbstractPolicyState",
    "AbstractActorCriticPolicy",
    "AbstractStatelessActorCriticPolicy",
    "AbstractStatefulActorCriticPolicy",
    "MLPActorCriticPolicy",
    "NCDEActorCriticPolicy",
]



from __future__ import annotations

import equinox as eqx
import optax
from jax import numpy as jnp
from jaxtyping import Array, ArrayLike, Bool, Float, Int, Key, Scalar, ScalarLike
from rich import progress, text
from tensorboardX import SummaryWriter

from lerax.env import AbstractEnvLike, AbstractEnvLikeState
from lerax.policy import AbstractPolicyState, AbstractStatefulPolicy
from lerax.utils import (
    callback_with_list_wrapper,
    callback_with_numpy_wrapper,
    callback_wrapper,
)


class EpisodeStats(eqx.Module):
    episode_return: Float[Array, ""]
    episode_length: Int[Array, ""]
    episode_done: Bool[Array, ""]

    @classmethod
    def initial(cls) -> EpisodeStats:
        return cls(
            jnp.array(0.0, dtype=float),
            jnp.array(0, dtype=int),
            jnp.array(False, dtype=bool),
        )

    def next(self, reward: Float[Array, ""], done: Bool[Array, ""]) -> EpisodeStats:
        return EpisodeStats(
            self.episode_return * (1.0 - self.episode_done.astype(float)) + reward,
            self.episode_length * (1 - self.episode_done.astype(int)) + 1,
            done,
        )


class StepCarry[PolicyType: AbstractStatefulPolicy](eqx.Module):
    env_state: AbstractEnvLikeState
    policy_state: AbstractPolicyState
    episode_stats: EpisodeStats

    @classmethod
    def initial(cls, env: AbstractEnvLike, policy: PolicyType, key: Key) -> StepCarry:
        env_state = env.initial(key=key)
        policy_state = policy.reset()
        return cls(env_state, policy_state, EpisodeStats.initial())


class IterationCarry[PolicyType: AbstractStatefulPolicy](eqx.Module):
    iteration_count: Int[Array, ""]
    step_carry: StepCarry[PolicyType]
    policy: PolicyType
    opt_state: optax.OptState


class JITSummaryWriter:
    """
    A wrapper around `tensorboardX.SummaryWriter` with a JIT compatible interface.
    """

    summary_writer: SummaryWriter

    def __init__(self, log_dir: str | None = None):
        self.summary_writer = SummaryWriter(log_dir=log_dir)

    def add_scalar(
        self,
        tag: str,
        scalar_value: ScalarLike,
        global_step: Int[ArrayLike, ""] | None = None,
        walltime: Float[ArrayLike, ""] | None = None,
    ):
        """
        Add a scalar value to the summary writer.
        """
        scalar_value = eqx.error_if(
            scalar_value,
            jnp.isnan(scalar_value) | jnp.isinf(scalar_value),
            "Scalar value cannot be NaN or Inf.",
        )
        callback_with_numpy_wrapper(self.summary_writer.add_scalar)(
            tag, scalar_value, global_step, walltime
        )

    def add_dict(
        self,
        scalars: dict[str, Scalar],
        prefix: str = "",
        *,
        global_step: Int[ArrayLike, ""] | None = None,
        walltime: Float[ArrayLike, ""] | None = None,
    ) -> None:
        """
        Log a dictionary of **scalar** values.
        """

        if prefix:
            scalars = {f"{prefix}/{k}": v for k, v in scalars.items()}

        for tag, value in scalars.items():
            self.add_scalar(tag, value, global_step=global_step, walltime=walltime)

    def log_episode_stats(
        self,
        episode_stats: EpisodeStats,
        *,
        first_step: Int[ArrayLike, ""],
    ) -> None:
        def log_finished(returns, lengths, dones, step_offset):
            for i, (_return, length, done) in enumerate(zip(returns, lengths, dones)):
                step = step_offset + i
                if done:
                    self.summary_writer.add_scalar(
                        "episode/return", _return, global_step=step
                    )
                    self.summary_writer.add_scalar(
                        "episode/length", length, global_step=step
                    )

        returns = episode_stats.episode_return
        lengths = episode_stats.episode_length
        dones = episode_stats.episode_done
        if returns.ndim == 1:
            returns = returns[None, :]
            lengths = lengths[None, :]
            dones = dones[None, :]

        num_envs, num_steps = returns.shape

        interleaved_returns = jnp.empty((num_steps * num_envs), dtype=returns.dtype)
        interleaved_lengths = jnp.empty((num_steps * num_envs), dtype=lengths.dtype)
        interleaved_dones = jnp.empty((num_steps * num_envs), dtype=dones.dtype)

        for env_idx in range(num_envs):
            interleaved_returns = interleaved_returns.at[env_idx::num_envs].set(
                returns[env_idx]
            )
            interleaved_lengths = interleaved_lengths.at[env_idx::num_envs].set(
                lengths[env_idx]
            )
            interleaved_dones = interleaved_dones.at[env_idx::num_envs].set(
                dones[env_idx]
            )

        callback_with_numpy_wrapper(log_finished)(
            interleaved_returns, interleaved_lengths, interleaved_dones, first_step
        )


def superscript_digit(digit: int) -> str:
    return "⁰¹²³⁴⁵⁶⁷⁸⁹"[digit % 10]


def superscript_int(i: int) -> str:
    return "".join(superscript_digit(int(c)) for c in str(i))


def suffixes(base: int):
    yield ""

    val = 1
    while True:
        yield f"×{base}{superscript_int(val)}"
        val += 1


def unit_and_suffix(value: float, base: int) -> tuple[float, str]:
    if base < 1:
        raise ValueError("base must be >= 1")

    unit, suffix = 1, ""
    for i, suffix in enumerate(suffixes(base)):
        unit = base**i
        if int(value) < unit * base:
            break

    return unit, suffix


class SpeedColumn(progress.ProgressColumn):
    """
    Renders human readable speed.

    https://github.com/NichtJens/rich/tree/master
    """

    def render(self, task: progress.Task) -> text.Text:
        """Show speed."""
        speed = task.finished_speed or task.speed

        if speed is None:
            return text.Text("", style="progress.percentage")
        unit, suffix = unit_and_suffix(speed, 2)
        data_speed = speed / unit
        return text.Text(f"{data_speed:.1f}{suffix} it/s", style="red")


class JITProgressBar:
    progress_bar: progress.Progress
    task: progress.TaskID

    def __init__(self, name: str, total: int | None, transient: bool = False):
        self.progress_bar = progress.Progress(
            progress.TextColumn("[progress.description]{task.description}"),
            progress.SpinnerColumn(finished_text="[green]✔"),
            progress.MofNCompleteColumn(),
            progress.BarColumn(bar_width=None),
            progress.TaskProgressColumn(),
            progress.TextColumn("["),
            progress.TimeElapsedColumn(),
            progress.TextColumn("<"),
            progress.TimeRemainingColumn(),
            progress.TextColumn("]"),
            SpeedColumn(),
            transient=transient,
        )
        self.task = self.progress_bar.add_task(f"[yellow]{name}", total=total)

    def start(self) -> None:
        callback_wrapper(self.progress_bar.start)()

    def stop(self) -> None:
        callback_wrapper(self.progress_bar.stop)()

    def update(
        self,
        total: Float[ArrayLike, ""] | None = None,
        completed: Float[ArrayLike, ""] | None = None,
        advance: Float[ArrayLike, ""] | None = None,
        description: str | None = None,
        visible: Bool[ArrayLike, ""] | None = None,
        refresh: Bool[ArrayLike, ""] = False,
    ) -> None:
        callback_with_list_wrapper(self.progress_bar.update)(
            self.task,
            total=total,
            completed=completed,
            advance=advance,
            description=description,
            visible=visible,
            refresh=refresh,
        )



from __future__ import annotations

import dataclasses

from jax import lax
from jax import numpy as jnp
from jaxtyping import Array, ArrayLike, Bool, Float, PyTree

from lerax.policy import AbstractPolicyState

from .base_buffer import AbstractBuffer


class RolloutBuffer[StateType: AbstractPolicyState, ActType, ObsType](AbstractBuffer):
    """
    RolloutBuffer used by on-policy algorithms.

    Designed for scans and JIT compilation.
    """

    observations: PyTree[ObsType]
    actions: PyTree[ActType]
    rewards: Float[Array, " *size"]
    dones: Bool[Array, " *size"]
    log_probs: Float[Array, " *size"]
    values: Float[Array, " *size"]
    returns: Float[Array, " *size"]
    advantages: Float[Array, " *size"]
    states: StateType

    def __init__(
        self,
        observations: PyTree[ObsType],
        actions: PyTree[ActType],
        rewards: Float[ArrayLike, " *size"],
        dones: Bool[ArrayLike, " *size"],
        log_probs: Float[ArrayLike, " *size"],
        values: Float[ArrayLike, " *size"],
        states: StateType,
        returns: Float[ArrayLike, " *size"] | None = None,
        advantages: Float[ArrayLike, " *size"] | None = None,
    ):
        """
        Initialize the RolloutBuffer with the given parameters.

        Returns and advantages can be provided, but if not, they will be filled with
        NaNs.
        """
        self.observations = observations
        self.actions = actions
        self.rewards = jnp.asarray(rewards, dtype=float)
        self.dones = jnp.asarray(dones, dtype=bool)
        self.log_probs = jnp.asarray(log_probs, dtype=float)
        self.values = jnp.asarray(values, dtype=float)
        self.states = states
        self.returns = (
            jnp.asarray(returns, dtype=float)
            if returns is not None
            else jnp.full_like(values, jnp.nan, dtype=float)
        )
        self.advantages = (
            jnp.asarray(advantages, dtype=float)
            if advantages is not None
            else jnp.full_like(values, jnp.nan, dtype=float)
        )

    def compute_returns_and_advantages(
        self,
        last_value: Float[ArrayLike, ""],
        gae_lambda: Float[ArrayLike, ""],
        gamma: Float[ArrayLike, ""],
    ) -> RolloutBuffer[StateType, ActType, ObsType]:
        last_value = jnp.asarray(last_value)
        gamma = jnp.asarray(gamma)
        gae_lambda = jnp.asarray(gae_lambda)

        next_values = jnp.concatenate([self.values[1:], last_value[None]], axis=0)
        next_non_terminals = 1.0 - self.dones.astype(float)
        deltas = self.rewards + gamma * next_values * next_non_terminals - self.values
        discounts = gamma * gae_lambda * next_non_terminals

        def scan_fn(
            carry: Float[Array, ""], x: tuple[Float[Array, ""], Float[Array, ""]]
        ) -> tuple[Float[Array, ""], Float[Array, ""]]:
            delta, discount = x
            advantage = delta + discount * carry
            return advantage, advantage

        _, advantages = lax.scan(
            scan_fn, jnp.array(0.0), (deltas, discounts), reverse=True
        )
        returns = advantages + self.values

        return dataclasses.replace(self, advantages=advantages, returns=returns)

    @property
    def shape(self) -> tuple[int, ...]:
        return self.rewards.shape



from __future__ import annotations

import time
from abc import abstractmethod
from typing import Literal, Sequence

import equinox as eqx
from jax import lax
from jax import random as jr
from jaxtyping import Array, Bool, Float, Key

from lerax.render import AbstractRenderer
from lerax.space import AbstractSpace
from lerax.utils import unstack_pytree


class AbstractEnvLikeState(eqx.Module):

    @property
    @abstractmethod
    def unwrapped(self) -> AbstractEnvState:
        """Return the unwrapped environment state"""


class AbstractEnvLike[StateType: AbstractEnvLikeState, ActType, ObsType](eqx.Module):
    """Base class for RL environments or wrappers that behave like environments"""

    name: eqx.AbstractVar[str]

    action_space: eqx.AbstractVar[AbstractSpace[ActType]]
    observation_space: eqx.AbstractVar[AbstractSpace[ObsType]]

    @abstractmethod
    def initial(self, *, key: Key) -> StateType:
        """Generate the initial state of the environment"""

    @abstractmethod
    def transition(self, state: StateType, action: ActType, *, key: Key) -> StateType:
        """Update the environment state given an action"""

    @abstractmethod
    def observation(self, state: StateType, *, key: Key) -> ObsType:
        """Generate an observation from the environment state"""

    @abstractmethod
    def reward(
        self, state: StateType, action: ActType, next_state: StateType, *, key: Key
    ) -> Float[Array, ""]:
        """Generate a reward from the environment state transition"""

    @abstractmethod
    def terminal(self, state: StateType, *, key: Key) -> Bool[Array, ""]:
        """Determine whether the environment state is terminal"""

    @abstractmethod
    def truncate(self, state: StateType) -> Bool[Array, ""]:
        """Determine whether the environment state is truncated"""

    @abstractmethod
    def state_info(self, state: StateType) -> dict:
        """Generate additional info from the environment state"""

    @abstractmethod
    def transition_info(
        self, state: StateType, action: ActType, next_state: StateType
    ) -> dict:
        """Generate additional info from the environment state transition"""

    @abstractmethod
    def default_renderer(self) -> AbstractRenderer:
        """Return the default renderer for the environment"""

    @abstractmethod
    def render(self, state: StateType, renderer: AbstractRenderer):
        """Render a frame from a state"""

    def render_states(
        self,
        states: Sequence[StateType],
        renderer: AbstractRenderer | Literal["auto"] = "auto",
        dt: float = 0.0,
    ):
        renderer = self.default_renderer() if renderer == "auto" else renderer
        renderer.open()
        for state in states:
            self.render(state, renderer)
            time.sleep(dt)
        renderer.close()

    def render_stacked(
        self,
        states: StateType,
        renderer: AbstractRenderer | Literal["auto"] = "auto",
        dt: float = 0.0,
    ):
        """Render multiple frames from stacked states"""
        self.render_states(unstack_pytree(states), renderer, dt)

    @property
    @abstractmethod
    def unwrapped(self) -> AbstractEnv:
        """Return the unwrapped environment"""

    @eqx.filter_jit
    def reset(self, *, key: Key) -> tuple[StateType, ObsType, dict]:
        """Wrap the functional logic into a Gym-like reset method"""
        initial_key, observation_key = jr.split(key, 2)
        state = self.initial(key=initial_key)
        observation = self.observation(state, key=observation_key)
        info = self.state_info(state)
        return state, observation, info

    @eqx.filter_jit
    def step(
        self, state: StateType, action: ActType, *, key: Key
    ) -> tuple[
        StateType, ObsType, Float[Array, ""], Bool[Array, ""], Bool[Array, ""], dict
    ]:
        """Wrap the functional logic into a Gym-like step method"""
        transition_key, reward_key, terminal_key, reset_key = jr.split(key, 4)

        next_state = self.transition(state, action, key=transition_key)
        reward = self.reward(state, action, next_state, key=reward_key)
        terminal = self.terminal(next_state, key=terminal_key)
        truncate = self.truncate(next_state)
        info = self.transition_info(state, action, next_state)

        state = lax.cond(
            terminal | truncate, lambda: self.initial(key=reset_key), lambda: next_state
        )
        observation = self.observation(state, key=key)

        return state, observation, reward, terminal, truncate, info


class AbstractEnvState(AbstractEnvLikeState):
    @property
    def unwrapped(self) -> AbstractEnvState:
        """Return the unwrapped environment state"""
        return self


class AbstractEnv[StateType: AbstractEnvState, ActType, ObsType](
    AbstractEnvLike[StateType, ActType, ObsType]
):
    """Base class for RL environments"""

    name: eqx.AbstractVar[str]

    action_space: eqx.AbstractVar[AbstractSpace[ActType]]
    observation_space: eqx.AbstractVar[AbstractSpace[ObsType]]

    @property
    def unwrapped(self) -> AbstractEnv[StateType, ActType, ObsType]:
        """Return the unwrapped environment"""
        return self



from __future__ import annotations

from abc import abstractmethod

import equinox as eqx
import optax
from jax import lax
from jax import numpy as jnp
from jax import random as jr
from jaxtyping import Key, Scalar

from lerax.buffer import RolloutBuffer
from lerax.env import AbstractEnvLike
from lerax.policy import AbstractActorCriticPolicy, AbstractStatefulActorCriticPolicy
from lerax.utils import filter_scan

from .base_algorithm import AbstractAlgorithm
from .utils import (
    EpisodeStats,
    IterationCarry,
    JITProgressBar,
    JITSummaryWriter,
    StepCarry,
)


class AbstractOnPolicyAlgorithm[PolicyType: AbstractActorCriticPolicy](
    AbstractAlgorithm[PolicyType]
):
    """
    Base class for on-policy algorithms.

    Generates rollouts using the current policy and estimates advantages and
    returns using GAE. Trains the policy using the collected rollouts.
    """

    optimizer: eqx.AbstractVar[optax.GradientTransformation]

    gae_lambda: eqx.AbstractVar[float]
    gamma: eqx.AbstractVar[float]

    num_envs: eqx.AbstractVar[int]
    num_steps: eqx.AbstractVar[int]
    batch_size: eqx.AbstractVar[int]

    def step(
        self,
        env: AbstractEnvLike,
        policy: AbstractStatefulActorCriticPolicy,
        carry: StepCarry,
        *,
        key: Key,
    ) -> tuple[StepCarry, RolloutBuffer]:
        (
            action_key,
            transition_key,
            observation_key,
            reward_key,
            terminal_key,
            bootstrap_key,
            reset_key,
        ) = jr.split(key, 7)

        observation = env.observation(carry.env_state, key=observation_key)
        next_policy_state, action, value, log_prob = policy.action_and_value(
            carry.policy_state, observation, key=action_key
        )

        next_env_state = env.transition(carry.env_state, action, key=transition_key)

        reward = env.reward(carry.env_state, action, next_env_state, key=reward_key)
        termination = env.terminal(next_env_state, key=terminal_key)
        truncation = env.truncate(next_env_state)
        done = termination | truncation
        next_episode_stats = carry.episode_stats.next(reward, done)

        # Bootstrap reward if truncated
        reward = lax.cond(
            truncation,
            lambda: reward
            + self.gamma
            * policy.value(
                next_policy_state, env.observation(next_env_state, key=bootstrap_key)
            )[1],
            lambda: reward,
        )

        # Reset environment if done
        next_env_state = lax.cond(
            done, lambda: env.initial(key=reset_key), lambda: next_env_state
        )

        # Reset policy state if done
        next_policy_state = lax.cond(
            done, lambda: policy.reset(), lambda: next_policy_state
        )

        return (
            StepCarry(next_env_state, next_policy_state, next_episode_stats),
            RolloutBuffer(
                observations=observation,
                actions=action,
                rewards=reward,
                dones=done,
                log_probs=log_prob,
                values=value,
                states=carry.policy_state,
            ),
        )

    def collect_rollout(
        self,
        env: AbstractEnvLike,
        policy: AbstractStatefulActorCriticPolicy,
        carry: StepCarry,
        key: Key,
    ) -> tuple[StepCarry, RolloutBuffer, EpisodeStats]:
        key, observation_key = jr.split(key, 2)

        def scan_step(
            carry: StepCarry, key: Key
        ) -> tuple[StepCarry, tuple[RolloutBuffer, EpisodeStats]]:
            carry, rollout = self.step(env, policy, carry, key=key)
            return carry, (rollout, carry.episode_stats)

        carry, (rollout_buffer, episode_stats) = filter_scan(
            scan_step, carry, jr.split(key, self.num_steps)
        )

        _, next_value = policy.value(
            carry.policy_state, env.observation(carry.env_state, key=observation_key)
        )
        rollout_buffer = rollout_buffer.compute_returns_and_advantages(
            next_value, self.gae_lambda, self.gamma
        )
        return carry, rollout_buffer, episode_stats

    @abstractmethod
    def train(
        self,
        policy: AbstractStatefulActorCriticPolicy,
        opt_state: optax.OptState,
        rollout_buffer: RolloutBuffer,
        *,
        key: Key,
    ) -> tuple[AbstractStatefulActorCriticPolicy, optax.OptState, dict[str, Scalar]]:
        """Train the policy using the rollout buffer."""

    def iteration(
        self,
        env: AbstractEnvLike,
        carry: IterationCarry,
        *,
        key: Key,
        progress_bar: JITProgressBar | None,
        tb_writer: JITSummaryWriter | None,
    ) -> IterationCarry:
        rollout_key, train_key = jr.split(key, 2)
        if self.num_envs == 1:
            step_carry, rollout_buffer, episode_stats = self.collect_rollout(
                env, carry.policy, carry.step_carry, key=rollout_key
            )
        else:
            step_carry, rollout_buffer, episode_stats = eqx.filter_vmap(
                self.collect_rollout, in_axes=(None, None, 0, 0)
            )(env, carry.policy, carry.step_carry, jr.split(rollout_key, self.num_envs))

        policy, opt_state, log = self.train(
            carry.policy, carry.opt_state, rollout_buffer, key=train_key
        )

        if progress_bar is not None:
            progress_bar.update(advance=self.num_steps * self.num_envs)

        if tb_writer is not None:
            first_step = carry.iteration_count * self.num_steps * self.num_envs
            final_step = first_step + self.num_steps * self.num_envs - 1
            log["learning_rate"] = optax.tree_utils.tree_get(
                opt_state, "learning_rate", jnp.nan
            )
            tb_writer.add_dict(log, prefix="train", global_step=final_step)
            tb_writer.log_episode_stats(episode_stats, first_step=first_step)

        return IterationCarry(carry.iteration_count + 1, step_carry, policy, opt_state)
